@@ -416,12 +416,15 @@ int Manage_subdevs(char *devname, int fd,
 	int lfd = -1;
 	int sysfd = -1;
 	int count = 0; /* number of actions taken */
+	struct mdinfo info;
+	int frozen = 0;
 
 	if (ioctl(fd, GET_ARRAY_INFO, &array)) {
 		fprintf(stderr, Name ": cannot get array info for %s\n",
 			devname);
-		return 1;
+		goto abort;
 	}
+	sysfs_init(&info, fd, 0);
 
 	/* array.size is only 32 bit and may be truncated.
 	 * So read from sysfs if possible, and record number of sectors
@@ -435,7 +438,7 @@ int Manage_subdevs(char *devname, int fd,
 	if (!tst) {
 		fprintf(stderr, Name ": unsupport array - version %d.%d\n",
 			array.major_version, array.minor_version);
-		return 1;
+		goto abort;
 	}
 
 	stb.st_rdev = 0;
@@ -445,7 +448,7 @@ int Manage_subdevs(char *devname, int fd,
 		char *dnprintable = dv->devname;
 		char *add_dev = dv->devname;
 		int err;
-		int re_add_failed = 0;
+		int array_failed;
 
 		next = dv->next;
 		jnext = 0;
@@ -457,9 +460,9 @@ int Manage_subdevs(char *devname, int fd,
 				fprintf(stderr, Name ": %s only meaningful "
 					"with -r, not -%c\n",
 					dv->devname, dv->disposition);
-				return 1;
+				goto abort;
 			}
-			for (; j < 1024 && remaining_disks > 0; j++) {
+			for (; j < MAX_DISKS && remaining_disks > 0; j++) {
 				unsigned dev;
 				disc.number = j;
 				if (ioctl(fd, GET_DISK_INFO, &disc))
@@ -490,9 +493,9 @@ int Manage_subdevs(char *devname, int fd,
 				fprintf(stderr, Name ": %s only meaningful "
 					"with -r of -f, not -%c\n",
 					dv->devname, dv->disposition);
-				return 1;
+				goto abort;
 			}
-			for (; j < 1024 && remaining_disks > 0; j++) {
+			for (; j < MAX_DISKS && remaining_disks > 0; j++) {
 				int sfd;
 				unsigned dev;
 				disc.number = j;
@@ -530,7 +533,7 @@ int Manage_subdevs(char *devname, int fd,
 			if (dv->disposition != 'a' || dv->re_add == 0) {
 				fprintf(stderr, Name ": 'missing' only meaningful "
 					"with --re-add\n");
-				return 1;
+				goto abort;
 			}
 			if (add_devlist == NULL)
 				add_devlist = conf_get_devs();
@@ -554,7 +557,7 @@ int Manage_subdevs(char *devname, int fd,
 				fprintf(stderr, Name ": %s only meaningful "
 					"with -r or -f, not -%c\n",
 					dv->devname, dv->disposition);
-				return 1;
+				goto abort;
 			}
 
 			sprintf(dname, "dev-%s", dv->devname);
@@ -576,7 +579,7 @@ int Manage_subdevs(char *devname, int fd,
 					fprintf(stderr, Name ": %s does not appear "
 						"to be a component of %s\n",
 						dv->devname, devname);
-					return 1;
+					goto abort;
 				}
 			}
 		} else {
@@ -595,7 +598,7 @@ int Manage_subdevs(char *devname, int fd,
 						dv->devname, strerror(errno));
 					if (tfd >= 0)
 						close(tfd);
-					return 1;
+					goto abort;
 				}
 				close(tfd);
 				tfd = -1;
@@ -604,21 +607,21 @@ int Manage_subdevs(char *devname, int fd,
 				fprintf(stderr, Name ": %s is not a "
 					"block device.\n",
 					dv->devname);
-				return 1;
+				goto abort;
 			}
 		}
 		switch(dv->disposition){
 		default:
 			fprintf(stderr, Name ": internal error - devmode[%s]=%d\n",
 				dv->devname, dv->disposition);
-			return 1;
+			goto abort;
 		case 'a':
 			/* add the device */
 			if (subarray) {
 				fprintf(stderr, Name ": Cannot add disks to a"
 					" \'member\' array, perform this"
 					" operation on the parent container\n");
-				return 1;
+				goto abort;
 			}
 			/* Make sure it isn't in use (in 2.6 or later) */
 			tfd = dev_open(add_dev, O_RDONLY|O_EXCL|O_DIRECT);
@@ -627,7 +630,13 @@ int Manage_subdevs(char *devname, int fd,
 			if (tfd < 0) {
 				fprintf(stderr, Name ": Cannot open %s: %s\n",
 					dv->devname, strerror(errno));
-				return 1;
+				goto abort;
+			}
+			if (!frozen) {
+				if (sysfs_freeze_array(&info) == 1)
+					frozen = 1;
+				else
+					frozen = -1;
 			}
 
 			st = dup_super(tst);
@@ -639,7 +648,7 @@ int Manage_subdevs(char *devname, int fd,
 				if (!get_dev_size(tfd, dv->devname, &ldsize)) {
 					st->ss->free_super(st);
 					close(tfd);
-					return 1;
+					goto abort;
 				}
 			} else if (!get_dev_size(tfd, NULL, &ldsize)) {
 				st->ss->free_super(st);
@@ -657,11 +666,11 @@ int Manage_subdevs(char *devname, int fd,
 						": %s is larger than %s can "
 						"effectively use.\n"
 						"       Add --force is you "
-						"really wan to add this device.\n",
+						"really want to add this device.\n",
 						add_dev, devname);
 					st->ss->free_super(st);
 					close(tfd);
-					return 1;
+					goto abort;
 				}
 				fprintf(stderr, Name
 					": %s is larger than %s can "
@@ -686,7 +695,7 @@ int Manage_subdevs(char *devname, int fd,
 
 				fprintf(stderr, Name ": hot add failed for %s: %s\n",
 					add_dev, strerror(errno));
-				return 1;
+				goto abort;
 			}
 
 			if (array.not_persistent == 0 || tst->ss->external) {
@@ -723,11 +732,17 @@ int Manage_subdevs(char *devname, int fd,
 					break;
 				}
 				/* FIXME this is a bad test to be using */
-				if (!tst->sb) {
+				if (!tst->sb &&
+				    dv->re_add) {
+					/* we are re-adding a device to a
+					 * completely dead array - have to depend
+					 * on kernel to check
+					 */
+				} else if (!tst->sb) {
 					close(tfd);
 					st->ss->free_super(st);
 					fprintf(stderr, Name ": cannot load array metadata from %s\n", devname);
-					return 1;
+					goto abort;
 				}
 
 				/* Make sure device is large enough */
@@ -740,19 +755,23 @@ int Manage_subdevs(char *devname, int fd,
 						continue;
 					fprintf(stderr, Name ": %s not large enough to join array\n",
 						dv->devname);
-					return 1;
+					goto abort;
 				}
 
 				/* Possibly this device was recently part of the array
 				 * and was temporarily removed, and is now being re-added.
 				 * If so, we can simply re-add it.
 				 */
-				tst->ss->uuid_from_super(tst, duuid);
 
 				if (st->sb) {
 					struct mdinfo mdi;
 					st->ss->getinfo_super(st, &mdi, NULL);
 					st->ss->uuid_from_super(st, ouuid);
+					if (tst->sb)
+						tst->ss->uuid_from_super(tst, duuid);
+					else
+						/* Assume uuid matches: kernel will check */
+						memcpy(duuid, ouuid, sizeof(ouuid));
 					if ((mdi.disk.state & (1<<MD_DISK_ACTIVE)) &&
 					    !(mdi.disk.state & (1<<MD_DISK_FAULTY)) &&
 					    memcmp(duuid, ouuid, sizeof(ouuid))==0) {
@@ -768,7 +787,7 @@ int Manage_subdevs(char *devname, int fd,
 						disc.number = mdi.disk.number;
 						if (ioctl(fd, GET_DISK_INFO, &disc) != 0
 						    || disc.major != 0 || disc.minor != 0
-						    || !enough_fd(fd))
+							)
 							goto skip_re_add;
 						disc.major = major(stb.st_rdev);
 						disc.minor = minor(stb.st_rdev);
@@ -789,7 +808,7 @@ int Manage_subdevs(char *devname, int fd,
 								fprintf(stderr, Name ": failed to open %s for"
 									" superblock update during re-add\n", dv->devname);
 								st->ss->free_super(st);
-								return 1;
+								goto abort;
 							}
 
 							if (dv->writemostly == 1)
@@ -812,7 +831,7 @@ int Manage_subdevs(char *devname, int fd,
 								fprintf(stderr, Name ": failed to update"
 									" superblock during re-add\n");
 								st->ss->free_super(st);
-								return 1;
+								goto abort;
 							}
 						}
 						/* don't even try if disk is marked as faulty */
@@ -830,11 +849,10 @@ int Manage_subdevs(char *devname, int fd,
 							st->ss->free_super(st);
 							if (add_dev != dv->devname)
 								continue;
-							return 1;
+							goto abort;
 						}
-					skip_re_add:
-						re_add_failed = 1;
 					}
+				skip_re_add:
 					st->ss->free_super(st);
 				}
 				if (add_dev != dv->devname) {
@@ -854,18 +872,36 @@ int Manage_subdevs(char *devname, int fd,
 					fprintf(stderr, Name
 						": --re-add for %s to %s is not possible\n",
 						dv->devname, devname);
-					return 1;
+					goto abort;
 				}
-				if (re_add_failed) {
-					fprintf(stderr, Name ": %s reports being an active member for %s, but a --re-add fails.\n",
-						dv->devname, devname);
-					fprintf(stderr, Name ": not performing --add as that would convert %s in to a spare.\n",
-						dv->devname);
-					fprintf(stderr, Name ": To make this a spare, use \"mdadm --zero-superblock %s\" first.\n",	
+				if (array.active_disks < array.raid_disks) {
+					char *avail = calloc(array.raid_disks, 1);
+					int d;
+					int found = 0;
+
+					for (d = 0; d < MAX_DISKS && found < array.active_disks; d++) {
+						disc.number = d;
+						if (ioctl(fd, GET_DISK_INFO, &disc))
+							continue;
+						if (disc.major == 0 && disc.minor == 0)
+							continue;
+						if (!(disc.state & (1<<MD_DISK_SYNC)))
+							continue;
+						avail[disc.raid_disk] = 1;
+						found++;
+					}
+					array_failed = !enough(array.level, array.raid_disks, 
+							       array.layout, 1, avail);
+				} else
+					array_failed = 0;
+				if (array_failed) {
+					fprintf(stderr, Name ": %s has failed so using --add cannot work and might destroy\n",
+						devname);
+					fprintf(stderr, Name ": data on %s.  You should stop the array and re-assemble it.\n",
 						dv->devname);
 					if (tfd >= 0)
 						close(tfd);
-					return 1;
+					goto abort;
 				}
 			} else {
 				/* non-persistent. Must ensure that new drive
@@ -876,7 +912,7 @@ int Manage_subdevs(char *devname, int fd,
 						dv->devname);
 					if (tfd >= 0)
 						close(tfd);
-					return 1;
+					goto abort;
 				}
 			}
 			/* committed to really trying this device now*/
@@ -911,11 +947,11 @@ int Manage_subdevs(char *devname, int fd,
 				if (tst->ss->add_to_super(tst, &disc, dfd,
 							  dv->devname)) {
 					close(dfd);
-					return 1;
+					goto abort;
 				}
 				if (tst->ss->write_init_super(tst)) {
 					close(dfd);
-					return 1;
+					goto abort;
 				}
 			} else if (dv->re_add) {
 				/*  this had better be raid1.
@@ -964,7 +1000,7 @@ int Manage_subdevs(char *devname, int fd,
 						" could not get exclusive access to container\n",
 						dv->devname);
 					tst->ss->free_super(tst);
-					return 1;
+					goto abort;
 				}
 
 				dfd = dev_open(dv->devname, O_RDWR | O_EXCL|O_DIRECT);
@@ -974,7 +1010,7 @@ int Manage_subdevs(char *devname, int fd,
 							  dv->devname)) {
 					close(dfd);
 					close(container_fd);
-					return 1;
+					goto abort;
 				}
 				if (tst->update_tail)
 					flush_metadata_updates(tst);
@@ -987,7 +1023,7 @@ int Manage_subdevs(char *devname, int fd,
 						dv->devname);
 					close(container_fd);
 					tst->ss->free_super(tst);
-					return 1;
+					goto abort;
 				}
 				sra->array.level = LEVEL_CONTAINER;
 				/* Need to set data_offset and component_size */
@@ -1003,7 +1039,7 @@ int Manage_subdevs(char *devname, int fd,
 						" failed for %s\n", dv->devname);
 					close(container_fd);
 					sysfs_free(sra);
-					return 1;
+					goto abort;
 				}
 				ping_monitor_by_id(devnum);
 				sysfs_free(sra);
@@ -1013,7 +1049,7 @@ int Manage_subdevs(char *devname, int fd,
 				if (ioctl(fd, ADD_NEW_DISK, &disc)) {
 					fprintf(stderr, Name ": add new device failed for %s as %d: %s\n",
 						dv->devname, j, strerror(errno));
-					return 1;
+					goto abort;
 				}
 			}
 			if (verbose >= 0)
@@ -1028,7 +1064,7 @@ int Manage_subdevs(char *devname, int fd,
 					" operation on the parent container\n");
 				if (sysfd >= 0)
 					close(sysfd);
-				return 1;
+				goto abort;
 			}
 			if (tst->ss->external) {
 				/* To remove a device from a container, we must
@@ -1048,7 +1084,7 @@ int Manage_subdevs(char *devname, int fd,
 						" to container - odd\n");
 					if (sysfd >= 0)
 						close(sysfd);
-					return 1;
+					goto abort;
 				}
 				/* in the detached case it is not possible to
 				 * check if we are the unique holder, so just
@@ -1065,7 +1101,7 @@ int Manage_subdevs(char *devname, int fd,
 						errno == EEXIST ? "still in use":
 						"not a member");
 					close(lfd);
-					return 1;
+					goto abort;
 				}
 			}
 			/* FIXME check that it is a current member */
@@ -1108,7 +1144,7 @@ int Manage_subdevs(char *devname, int fd,
 					strerror(errno));
 				if (lfd >= 0)
 					close(lfd);
-				return 1;
+				goto abort;
 			}
 			if (tst->ss->external) {
 				/*
@@ -1121,7 +1157,7 @@ int Manage_subdevs(char *devname, int fd,
 
 				if (!name) {
 					fprintf(stderr, Name ": unable to get container name\n");
-					return 1;
+					goto abort;
 				}
 
 				ping_manager(name);
@@ -1144,7 +1180,7 @@ int Manage_subdevs(char *devname, int fd,
 					dnprintable, strerror(errno));
 				if (sysfd >= 0)
 					close(sysfd);
-				return 1;
+				goto abort;
 			}
 			if (sysfd >= 0)
 				close(sysfd);
@@ -1156,9 +1192,16 @@ int Manage_subdevs(char *devname, int fd,
 			break;
 		}
 	}
+	if (frozen > 0)
+		sysfs_set_str(&info, NULL, "sync_action","idle");
 	if (test && count == 0)
 		return 2;
 	return 0;
+
+abort:
+	if (frozen > 0)
+		sysfs_set_str(&info, NULL, "sync_action","idle");
+	return 1;
 }
 
 int autodetect(void)
